/** Return the average of a stats entity. */
function AVG(t)
{
  return (t[0] ? t[1] / t[0] : 0);
}

/** Return the standard deviation of a stats entity. */
function RMS(t)
{
  if (t[0] == 0)
    return 0;

  var avg = AVG(t);
  var v = t[2] / t[0] - avg*avg;

  if (v > 0)
    return Math.sqrt(v);
  else
    return 0;
}

/** Functions for retrieving different stats entities. */
function NREQ(svc) { return svc.nreq; }
function TIME_AVG(svc) { return AVG(svc.total[0]); }
function TIME_RMS(svc) { return RMS(svc.total[0]); }
function SIZE_AVG(svc) { return AVG(svc.total[1]); }
function SIZE_RMS(svc) { return RMS(svc.total[1]); }
var YLABELS = { "NREQ": "Count",
		"TIME_AVG": "Milliseconds",
		"TIME_RMS": "Milliseconds",
		"SIZE_AVG": "Kilobytes",
		"SIZE_RMS": "Kilobytes" };

/** General parameters. */
var w = 700, h = 350;
var _data = [];
var _days = {};
var _hide = {};
var _qty = NREQ;
var _ylabel = YLABELS["NREQ"];
var _scale = "linear";
var _stats_tbl = null;
var _stats_qty = null;
var _stats_body = null;

/** XHR request failure handler. */
function onXHRFail(o)
{
  var headers = []; for (var h in o) { headers.push(h); }
  if (console && console.error)
  {
    console.error(_dump(headers));
    console.error("failed to load data for month " + o.argument
		  + ": http status " + o.status + " (" + o.statusText + ")");

    for (var i in o.getResponseHeader)
      console.error("response header " + i + ": " + o.getResponseHeader[i]);

    console.error("response text: " + o.responseText);
  }
}

/** Handler for loading summary data. */
function onLoadSummary(o)
{
  if (o.status != 200)
    return onXHRFail(o);

  parseSummary(o.responseText, o.argument);
  refreshPlot();
  refreshTable();
}

/** Parse summary JSON result. Reprocess the compressed JSON to
    expand strings and abbreviated objects. At this point strings
    are real JS objects, so expanding string references does not
    add to memory use. Also convert objects from arrays to real
    JS objects by expanding column names. This makes downstream
    code more readable. */
function parseSummary(json, month)
{
  var data = eval("(" + json + ")");
  var strs = data.strings;
  var cols = data.columns;
  var tbins = data.data;

  /* Process time bin objects { bin: name, data: [svcs] }. Assign
     the daily data to the corresponding object in _days; throw
     away anything we don't have in _days: undesired dates and
     the "TOTAL" bin. */
  for (tb in tbins)
  {
    if (! (tbins[tb].bin in _days))
      continue;

    _days[tbins[tb].bin].data = tbins[tb].items.map(function(svc)
    {
      /* Convert the day data to an object. Expand array elements
         to named object attributes. Expand strings in the stats
         arrays and convert them to objects we can use with _qty(). */
      var obj = {};
      for (var i = 0, e = cols.length; i < e; ++i)
      {
        if (cols[i] == "service")
          svc[i] = data.strings[svc[i]];

        else if (cols[i] != "nreq" && cols[i] != "total")
        {
          var arr = svc[i];
          for (var j in arr)
            arr[j] = { name: data.strings[arr[j][0]],
		       nreq: arr[j][1], total: arr[j].slice(2) };
        }

        obj[cols[i]] = svc[i];
      }

      return obj;
    });
  }
}

/** Refresh the plot area to match updated data / selection. */
function refreshPlot()
{
  var fig = $("fig");
  fig.innerHTML = "";

  var xmax = 0;
  var ymax = 0;
  var svcs = { ALL: 0 };
  var d = { ALL: _data.map(function(i) { return { x: i.day, y: 0 }; }) };
  for (var daynr in _data)
  {
    var day = _data[daynr];
    if (day.day > xmax)
      xmax = day.day;

    for (var item in day.data)
    {
      var svc = day.data[item];
      if (! (svc.service in d))
      {
        d[svc.service] = _data.map(function(i) { return { x: i.day, y: 0 }; });
        svcs[svc.service] = 0;
      }

      var val = _qty(svc);
      d[svc.service][daynr].y = val;
      svcs[svc.service] += val;

      if (! _hide[svc.service] && val > ymax)
        ymax = val;
    }
  }

  if (_scale == "log")
  {
    for (var svc in d)
      for (var daynr in d[svc])
        if (d[svc][daynr].y == 0)
	  d[svc][daynr].y = 0.999;
  }

  /* Sizing and scales. */
  var x = pv.Scale.linear(0, xmax).range(0, w);
  var y = pv.Scale[_scale](_scale == "log" ? 1 : 0, ymax).range(0, h);
  var xticks = x.ticks(7);
  var yticks = y.ticks(10);

  /* The root panel. */
  var vis = new pv.Panel();
  vis.canvas("fig")
     .width(w)
     .height(h)
     .bottom(20)
     .left(70)
     .right(100)
     .top(5);

  /* X/Y grid. */
  vis.add(pv.Rule).data(yticks).bottom(y).strokeStyle("#ddd");
  vis.add(pv.Rule).data(xticks).bottom(3).left(x).strokeStyle("#ddd");

  /* Y-axis and ticks. */
  vis.add(pv.Rule)
     .data(yticks)
     .left(-2)
     .bottom(y)
     .width(5)
     .strokeStyle("#333")
     .anchor("left")
     .add(pv.Label)
     .text(_scale != "log" ? y.tickFormat :
           function(i) { return i.toPrecision(2); })
     .visible(_scale != "log" ? true :
              function(i) { return i.toPrecision(1).match(/^1e/) ? 1 : 0; });

  vis.add(pv.Label)
     .data([_ylabel])
     .left(-55)
     .bottom(h/2)
     .font("10pt Arial")
     .textAlign("center")
     .textAngle(-Math.PI/2);

  /* X-axis and ticks. */
  vis.add(pv.Rule)
     .data(xticks)
     .bottom(-2)
     .left(x)
     .height(5)
     .strokeStyle("#333")
     .anchor("bottom")
     .add(pv.Label)
     .bottom(-5)
     .height(5)
     .text(function(i) { return _data[d.ALL[i].x].date; });

  /* The area. */
  var svcnames = [];
  for (s in svcs) { svcnames.push(s); }
  svcnames.sort(function(a, b) { return svcs[b] - svcs[a]; });

  var c = pv.Colors.category20(svcnames);
  for (i = svcnames.length-1, e = 0; i >= e; --i)
  {
    var s = svcnames[i];
    vis.add(pv.Line)
       .data(d[s])
       .lineWidth(s == "ALL" ? 3 : 1)
       .strokeStyle(c(s)) /* function(i) { return _scale == "log" && i.y < 1 ? "rgba(255,255,255,0)" : c(s); }) */
       .left(function(i) { return x(i.x); })
       .bottom(function(i) { return y(i.y); })
       .visible(! _hide[s]);
  }

  /* Legend. */
  vis.add(pv.Dot)
     .data(svcnames)
     .right(-6)
     .top(function() { return 5 + this.index * 15; })
     .size(8)
     .strokeStyle(function(i) { return c(i); })
     .fillStyle(function(i) { return _hide[i] ? "#fff" : c(i); })
     .cursor("pointer")
     .event("mousedown", toggle)
     .anchor("right")
     .add(pv.Label)
     .cursor("pointer")
     .events("all")
     .event("mousedown", function() { toggle(this.text()); });

  vis.render();
}

/** Refresh the table to match the updated data / selection. */
function refreshTable()
{
  var ndays = _data.length;
  var dayopt = $('day-opt').value;
  var detail = $('detail').value;
  for (var daynr in _data)
  {
    var day = _data[daynr];
    var mainrow = _stats_body.childNodes[ndays - daynr - 1];
    var detailrow = mainrow.lastChild;
    var total = 1;
    var all = null;
    var svcs = [];
    for (var item in day.data)
    {
      var svc = day.data[item];
      if (svc.service == "ALL")
      {
        all = svc;
	total = _qty(svc);
        mainrow.childNodes[1].innerHTML = thousands(total);
      }
      else if (! _hide[svc.service])
        svcs.push({ qty: _qty(svc), data: svc });
    }

    var sub = "";
    if (dayopt == 'summary')
      ;
    else if (dayopt == 'service')
    {
      /* FIXME: Service detail */
      svcs.sort(function(a, b) { return b.qty - a.qty; });
      for (var item in svcs)
      {
	var row = svcs[item];
	sub += "<div><span class='qty'>"
	    + thousands(row.qty)
	    + "</span><span class='pct'>"
	    + thousands(sprintf("%.1f%%", row.qty / total * 100.))
	    + "</span><span class='info'>"
	    + row.data.service
	    + "</span></div>";
      }
    }
    else if (all)
    {
      var items = (all[detail] || []).map(function(a) {
	      return { val: a.name, qty: _qty(a) }; });
      items.sort(function(a, b) { return b.qty - a.qty; });
      for (var i in items)
	sub += "<div><span class='qty'>"
	    + thousands(items[i].qty)
	    + "</span><span class='pct'>"
	    + thousands(sprintf("%.1f%%", items[i].qty / total * 100.))
	    + "</span><span class='info'>"
	    + items[i].val
	    + "</span></div>";
    }

    detailrow.innerHTML = sub;
    detailrow.style.display = (dayopt == 'summary' ? 'none' : '');
  }
}

function toggle(svc)
{
  _hide[svc] = ! _hide[svc];
  refreshPlot();
  refreshTable();
}

function init()
{
  _stats_tbl = $('stats');
  _stats_qty = $('stats-qty');
  _stats_body = $('stats-body');
  _stats_qty.innerHTML = YLABELS["NREQ"];

  var tab = "";
  var day = 0;
  var months = {};
  var now = parseInt((new Date()).getTime() / 1000);
  var days = timeseries('day', now - 120 * 86400, now);
  for (var di = 0, de = days.length; di < de; ++di)
  {
    var d = days[di];
    d.day = day++;
    d.tm = new Date(d.start * 1000);
    d.date = sprintf("%04d-%02d-%02d",
                     d.tm.getUTCFullYear(),
                     d.tm.getUTCMonth() + 1,
                     d.tm.getUTCDate());
    d.month = sprintf("%04d%02d",
                      d.tm.getUTCFullYear(),
                      d.tm.getUTCMonth() + 1);
    d.data = [];
    _data.push(d);
    _days[d.date] = d;

    if (! (d.month in months))
    {
      YAHOO.util.Connect.asyncRequest
        ("GET", "stats/stats-" + d.month + "-json/summary.txt",
         { argument: d.month, success: onLoadSummary, failure: onXHRFail });
      months[d.month] = 1;
    }

    tab = "<div class='summary " + ((d.day & 1) ? "odd" : "even")
	+ "'><span class='day'>" + d.date + "</span>"
	+ "<span class='qty'>...</span>"
	+ "<span class='pct'>100%</span>"
	+ "<span class='info'>&nbsp;</span>"
	+ "<div class='nested detail'></div></div>"
	+ tab;
  }

  _stats_body.innerHTML = tab;

  $('qty').onchange = function() {
    _qty = eval(this.value);
    _ylabel = YLABELS[this.value];
    _stats_qty.innerHTML = _ylabel;
    refreshPlot();
    refreshTable();
    return false;
  };

  $('scale').onchange = function() {
    _scale = this.value;
    refreshPlot();
    refreshTable();
    return false;
  };

  $('day-opt').onchange = $('detail').onchange = function() {
    refreshTable();
    return false;
  };

  refreshPlot();
  refreshTable();
}

init();
