/** Utility to dump an object @a o as a string. */
function _dump(o)
{
  if (typeof(o) == 'object')
  {
    var result = "{ ";
    var joiner = "";
    for (i in o)
    {
      result += joiner + i + ": " + _dump(o[i]);
      joiner = ", ";
    }
    return result + " }";
  }
  else if (typeof(o) == typeof(""))
    return "\"" + o + "\"";
  else
    return "" + o;
}

/** Utility function to look up a document element or elements by
    name.  The arguments are either a single string, or an array of
    strings.  For the former returns a document element by that name,
    for the latter returns an array of elements, one for each name. */
function $()
{
  var elements = new Array();
  for (var i = 0; i < arguments.length; i++)
  {
    var element = arguments[i];
    if (typeof element == 'string')
      element = document.getElementById(element);
    if (arguments.length == 1)
      return element;
    elements.push(element);
  }
  return elements;
}

/** Utility function to look up a document element by type.  Returns
    elements of type @a type that are children of @a el. */
function $$(el, type)
{
  return (el || document).getElementsByTagName(type);
}

/** Utility to make string replacements.

    For each ${var} in @a str looks for replacements in @a opts, which
    may be a dictionary or a function.  If it is a dictionary and the
    value of key 'var' is a string or a number, the replacement is
    made.  If @a opts is a function, it's invoked with the name 'var'
    for each replacement.

    Returns the string after all possible replacements have been made. */
function _(str, opts)
{
  return str.replace(/\$\{([^{}]*)\}/g,
		     function (a, b) {
		       if (typeof opts === 'function')
			 return opts(b);
		       var r = opts[b];
		       return typeof r === 'string' || typeof r === 'number' ? r : a;
		     });
}

/** Utility to generate time series. */
function timeseries(span, start, end)
{
  var series = [];
  if (span == "hour")
  {
    var low = parseInt(start / 3600);
    var high = Math.max(low+1, parseInt((end+3599) / 3600));
    for (var t = low; t < high; ++t)
      series.push({ start: t * 3600, end: (t+1) * 3600 });
  }
  else if (span == "day")
  {
    var low = parseInt(start / 86400);
    var high = Math.max(low+1, parseInt((end+86399) / 86400));
    for (var t = low; t < high; ++t)
      series.push({ start: t * 86400, end: (t+1) * 86400 });
  }
  else if (span == "week")
  {
    var wday = (new Date(start * 1000)).getUTCDay();
    wday = (wday == 0 ? 6 : wday++);
    var low = parseInt(start / 86400) - wday;
    var high = Math.max(low+1, parseInt((end+86399) / 86400));
    for (var t = low; t < high; t += 7)
      series.push({ start: t * 86400, end: (t+7) * 86400 });
  }
  else if (span == "month")
  {
    var limit = parseInt((end+86399)/86400) * 86400;
    var t = new Date(start * 1000);
    var year = t.getUTCFullYear();
    var month = t.getUTCMonth();
    var day = parseInt(Date.UTC(year, month, 1, 0, 0, 0) / 1000);
    do
    {
      var low = day;
      if (++month > 12)
      {
        month = 1;
        year++;
      }

      day = parseInt(Date.UTC(year, month, 1, 0, 0, 0) / 1000);
      series.push({ start: low, end: day });
    }
    while (day < limit);
  }
  else if (span == "year")
  {
    var limit = parseInt((end+86399)/86400) * 86400;
    var year = (new Date(start * 1000)).getUTCFullYear();
    var day = parseInt(Date.UTC(year, 1, 1, 0, 0, 0) / 1000);
    do
    {
      var low = day;
      day = Date.UTC(++year, 1, 1, 0, 0, 0);
      series.push({ start: low, end: day });
    }
    while (day < limit);
  }
  else
    console.error("timeseries(): bad span '" + span + "'");

  return series;
}

/** Utility to render numbers with thousand separators. */
var RE_THOUSANDS = /(\d)(\d{3}($|\D))/;
var RE_DECIMAL = /^(.*?)(\.[^.]*)?$/;
function thousands(val)
{
  var r = RE_DECIMAL.exec(val);
  var prefix = r[1];
  var suffix = r[2];
  if (suffix)
    suffix = suffix.substring(0, 3);
  else
    suffix = "";

  while (true)
  {
    var t = prefix.replace(RE_THOUSANDS, "$1'$2")
    if (t == prefix)
      return t + suffix;
    else
      prefix = t;
  }
  return prefix + suffix;
}
